// Copyright 2023 Northern.tech AS
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

#include <gtest/gtest.h>

#include <common/log.hpp>
#include <common/setup.hpp>

using namespace testing;

class GlobalTestEnvironment : public ::testing::Environment {
public:
	void SetUp() override {
		mender::common::setup::GlobalSetup();
		mender::common::log::SetLevel(mender::common::log::LogLevel::Trace);
	}
};

int main(int argc, char **argv) {
	printf("Running mender_test::main\n");

	::testing::AddGlobalTestEnvironment(new GlobalTestEnvironment);
	::testing::InitGoogleTest(&argc, argv);

	return RUN_ALL_TESTS();
}
